#ifndef _NET_MPLS_H_
#define _NET_MPLS_H_

#include <asm/atomic.h>
#include <linux/mpls.h>
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>

struct fib_result;
struct rtable;

extern int mpls_debug;
#define MPLS_DEBUG(x)		if(mpls_debug) printk x

#define RADIX_INIT(head)						\
{									\
  (head)->RdX_root = NULL;						\
}

#define RADIX_ENTRY(type,bits)						\
struct {								\
  struct type *RdX_next[(0x01 << bits)];				\
}

#define RADIX_HEAD(name,type)						\
struct name {								\
  struct type *RdX_root;						\
};

#define RADIX_INSERT(head,type,field,bits,tdef,key,klen,result)		\
{									\
  struct type **RdX_node = &((head)->RdX_root);				\
  tdef RdX_mask;							\
  int RdX_index;							\
  int RdX_size = sizeof(tdef)*8;					\
  char RdX_status = 0;							\
  int RdX_count = 0;							\
									\
  if(klen % bits) result = -1;						\
  else {								\
    RdX_mask = (0x01 << bits)-1;					\
    while((RdX_count*bits) < klen) {					\
      RdX_status = 1;							\
      if(!(*RdX_node)) {						\
         RdX_status = 2;						\
         (*RdX_node) = (struct type *)kmalloc(sizeof(struct type),	\
           GFP_KERNEL);							\
         if(!(*RdX_node)) {						\
           result = -2;							\
           break;							\
         }								\
         RdX_status = 3;						\
         memset((*RdX_node),0,sizeof(struct type));			\
         RdX_status = 4;						\
      }									\
      RdX_index = (key >> (RdX_size - bits - (RdX_count * bits))) & RdX_mask; \
      RdX_status = 5;							\
      RdX_node = &((*RdX_node)->field.RdX_next[RdX_index]);		\
      RdX_status = 6;							\
      RdX_count++;							\
      result = 0;							\
    }									\
    if(result != -2 && (!(*RdX_node))) {				\
      (*RdX_node) = (struct type *)kmalloc(sizeof(struct type),		\
        GFP_KERNEL);							\
      if(!(*RdX_node)) {						\
        result = -2;							\
      } else {								\
        memset((*RdX_node),0,sizeof(struct type));			\
      }									\
    }									\
  }									\
}

#define RADIX_DO(head,type,field,bits,tdef,key,klen,flm,elem,result,RxD_set) \
{									\
  struct type *RdX_node = (head)->RdX_root;				\
  tdef RdX_mask;							\
  int RdX_index;							\
  int RdX_size = sizeof(tdef)*8;					\
  int RdX_count = 0;							\
									\
  if(klen % bits) result = -1;						\
  else {								\
    RdX_mask = (0x01 << bits)-1;					\
    while((RdX_count*bits) < klen) {					\
      if(!RdX_node) {							\
        result = -2;							\
        break;								\
      } else {								\
        RdX_index = (key >> (RdX_size - bits - (RdX_count * bits))) & RdX_mask;\
        RdX_node = RdX_node->field.RdX_next[RdX_index];			\
      }									\
      RdX_count++;							\
    }									\
    if(!RdX_node) {							\
      result = -3;							\
    } else {								\
      if(RxD_set) {							\
        RdX_node->flm = elem;						\
      } else {								\
        elem = RdX_node->flm;						\
      }									\
      result = 0;							\
    }									\
  }									\
}

#define RADIX_SET(head,type,field,bits,tdef,key,klen,flm,elem,result)	\
        RADIX_DO(head,type,field,bits,tdef,key,klen,flm,elem,result,1)
#define RADIX_GET(head,type,field,bits,tdef,key,klen,flm,elem,result)	\
        RADIX_DO(head,type,field,bits,tdef,key,klen,flm,elem,result,0)

#define RADIX_VISIT_ALL(head,type,field,bits,visit,extra)		\
{									\
  int RdX_node_count = (0x1 << bits);					\
  struct type *RdX_stack[RdX_node_count];				\
  int RdX_stack_count[RdX_node_count];					\
  int RdX_stack_depth = 0;						\
  struct type *RdX_node;						\
  int RdX_done = 0;							\
  int R = 0;								\
									\
  if((RdX_node = (head)->RdX_root)) {					\
    while(!RdX_done) {							\
      if(RdX_node->field.RdX_next[R]) {					\
        RdX_stack[RdX_stack_depth] = RdX_node;				\
        RdX_stack_count[RdX_stack_depth] = R + 1;			\
        RdX_stack_depth++;						\
        RdX_node = RdX_node->field.RdX_next[R];				\
        R = 0;								\
      } else {								\
        R++;								\
        if(R >= RdX_node_count) {					\
          if(visit(RdX_node,extra)) {					\
            RdX_done = 1;						\
            break;							\
          }								\
          RdX_stack_depth--;						\
          if(RdX_stack_depth >= 0) {					\
            RdX_node = RdX_stack[RdX_stack_depth];			\
            R = RdX_stack_count[RdX_stack_depth];			\
          } else {							\
            RdX_done = 1;						\
          }								\
        }								\
      }									\
    }									\
  }									\
}

#define MPLS_TREE_BITS		4
#define MPLS_TREE_DEPTH		32

#define MPLS_RESULT_SUCCESS	0
#define MPLS_RESULT_RECURSE	1
#define MPLS_RESULT_DROP	2
#define MPLS_RESULT_DLV		3
#define MPLS_RESULT_FWD		4

struct mpls_skb_parm {
  unsigned int gap;
};

#define MPLSCB(skb) ((struct mpls_skb_parm*)((skb)->cb))

struct mpls_push_data {
  unsigned int label:20;
  unsigned int ttl:8;
  unsigned int exp:3;
  unsigned int bos:1;
  unsigned char flag;
  char popped_bos;
};

struct mpls_instruction {
  unsigned short mi_opcode;
  void *mi_data;
};

struct mpls_nfmark_fwd_info {
  struct mpls_out_info *nfi_moi[MPLS_NFMARK_NUM];
  unsigned short nfi_mask;
};

struct mpls_dsmark_fwd_info {
  struct mpls_out_info *dfi_moi[MPLS_DSMARK_NUM];
  unsigned char dfi_mask;
};

struct mpls_tcindex_fwd_info {
  struct mpls_out_info *tfi_moi[MPLS_TCINDEX_NUM];
  unsigned short tfi_mask;
};

struct mpls_exp_fwd_info {
  struct mpls_out_info *efi_moi[MPLS_EXP_NUM];
};

struct mpls_exp2dsmark_info {
  unsigned char e2d[MPLS_EXP_NUM];
};

struct mpls_exp2tcindex_info {
  unsigned short e2t[MPLS_EXP_NUM];
};

struct mpls_tcindex2exp_info {
  unsigned char t2e_mask;
  unsigned char t2e[MPLS_TCINDEX_NUM];
};

struct mpls_dsmark2exp_info {
  unsigned char d2e_mask;
  unsigned char d2e[MPLS_DSMARK_NUM];
};

struct mpls_nfmark2exp_info {
  unsigned char n2e_mask;
  unsigned char n2e[MPLS_NFMARK_NUM];
};

struct mpls_in_info {
  atomic_t __refcnt;
  struct list_head dev_entry;
  struct list_head moi_entry;
  struct mpls_instruction mii_instruction[MPLS_NUM_OPS];
  struct mpls_label mii_label;
  unsigned int mii_key;
  unsigned int mii_age;
  unsigned short mii_proto;
  unsigned short mii_labelspace;
  unsigned short mii_instruction_length;
  unsigned long mii_packets;
  unsigned long mii_bytes;
  unsigned long mii_drops;
};

struct mpls_in_info_node {
  RADIX_ENTRY(mpls_in_info_node,MPLS_TREE_BITS) next;
  struct mpls_in_info *mii;
};
RADIX_HEAD(mpls_in_info_tree,mpls_in_info_node);

extern struct mpls_in_info_tree mii_tree;

struct mpls_out_info {
  atomic_t __refcnt;
  struct notifier_block *moi_notifier_list;
  struct list_head list_out;
  struct list_head list_in;
  struct list_head dev_entry;
  struct list_head moi_entry;
  struct mpls_instruction moi_instruction[MPLS_NUM_OPS];
  int moi_ifindex;
  unsigned int moi_key;
  unsigned int moi_age;
  unsigned short moi_instruction_length;
  unsigned short moi_mtu_limit;
  unsigned short moi_mtu;
  unsigned char moi_propogate_ttl;
  unsigned long moi_packets;
  unsigned long moi_bytes;
  unsigned long moi_drops;
};

struct mpls_out_info_node {
  RADIX_ENTRY(mpls_out_info_node,MPLS_TREE_BITS) next;
  struct mpls_out_info *moi;
};
RADIX_HEAD(mpls_out_info_tree,mpls_out_info_node);

struct mpls_dst {
  atomic_t __refcnt;
  struct dst_entry *md_dst;
  struct sockaddr md_nh;
};

struct mpls_tunnel_private {
  struct mpls_out_info *mtp_moi;
  struct net_device *mtp_dev;
  struct sockaddr mtp_dest;
  struct mpls_tunnel_private* next;
  struct net_device_stats stat;
};

struct mpls_interface {
  struct list_head list_out;
  struct list_head list_in;
  int labelspace;
};

struct mpls_fwd_block {
  struct notifier_block notifier_block;
  struct mpls_out_info *owner;
  struct mpls_out_info *fwd;
};

struct mpls_ops {
 int (*fn)(struct sk_buff** skb,struct mpls_in_info *mii,
  struct mpls_out_info **moi, struct mpls_push_data *mpr, void *data);
 char *msg;
 int extra;
};

extern struct mpls_out_info_tree moi_tree;

extern void __mpls_del_in_label(struct mpls_in_info*);
extern void __mpls_del_out_label(struct mpls_out_info*);

extern void mpls_skb_dump(struct sk_buff* sk);
extern void mpls_dst_hold(struct mpls_dst *md);
extern void mpls_dst_release(struct mpls_dst *md);
extern void mpls_label_hold(struct mpls_label *ml);
extern void mpls_label_release(struct mpls_label *ml);
extern void mpls_in_info_hold(struct mpls_in_info *mii);
extern void mpls_in_info_release(struct mpls_in_info *mii);
extern void mpls_out_info_hold(struct mpls_out_info *moi);
extern void mpls_out_info_release(struct mpls_out_info *moi);

extern int mpls_add_in_label(struct mpls_in_label_req *in);
extern int mpls_get_in_label(struct mpls_in_label_req *in);
extern int mpls_del_in_label(struct mpls_in_label_req *in);
extern int mpls_attach_in2out(struct mpls_xconnect_req *req);
extern int mpls_detach_in2out(struct mpls_xconnect_req *req);
extern int mpls_get_in2out(struct mpls_xconnect_req *req);
extern int mpls_set_in_label_proto(struct mpls_in_label_req *in);

extern int mpls_info_default_in_instruction(struct mpls_in_info *mii);
extern int mpls_set_default_out_instruction(struct mpls_out_info*);
extern int mpls_set_in_label_instructions(struct mpls_instruction_req *in);
extern int mpls_get_in_label_instructions(struct mpls_instruction_req *in);
extern int mpls_set_out_label_instructions(struct mpls_instruction_req *out);
extern int mpls_get_out_label_instructions(struct mpls_instruction_req *out);
extern int mpls_add_out_label(struct mpls_out_label_req *out);
extern int mpls_get_out_label(struct mpls_out_label_req *out);
extern int mpls_del_out_label(struct mpls_out_label_req *out);
extern int mpls_set_out_label_mtu(struct mpls_out_label_req *out);

extern struct mpls_in_info *mpls_get_mii(unsigned int key);
extern struct mpls_in_info *mpls_get_mii_by_label(struct mpls_push_data *mpr,
  struct mpls_label *label, int labelspace);
extern struct mpls_out_info *mpls_get_moi(unsigned int key);
extern int mpls_insert_mii(unsigned int key, struct mpls_in_info *mii);
extern int mpls_insert_moi(unsigned int key, struct mpls_out_info *moi);

extern void mpls_flush_in_tree(void);
extern void mpls_flush_out_tree(void);

extern int mpls_ioctl(unsigned int cmd,void *arg);
extern int mpls_output(struct sk_buff *skb);
extern unsigned int mpls_label2key(int,struct mpls_label*);
extern int mpls_rcv(struct sk_buff*,struct net_device*,struct packet_type*);

extern int mpls_out_label_info(char*,char**,off_t,int);
extern int mpls_in_label_info(char*,char**,off_t,int);
extern int mpls_labelspace_info(char*,char**,off_t,int);

extern u32 mpls_opcode_pop(struct sk_buff *skb);
extern char mpls_opcode_push(struct sk_buff**,struct mpls_label*,struct mpls_push_data*);
extern int mpls_opcode_peek(struct sk_buff*,struct mpls_push_data*);
extern int mpls_input(struct sk_buff*,struct net_device*,struct packet_type*,struct mpls_label*,struct mpls_push_data*,int);
extern void mpls_finish(struct sk_buff*);
extern int mpls_output2(struct sk_buff*,struct mpls_out_info*,struct mpls_push_data*);

extern int mpls_get_labelspace(struct mpls_labelspace_req *req);
extern int mpls_set_labelspace(struct mpls_labelspace_req *req);

extern void mpls_print_label(struct mpls_label*);
extern void mpls_instruction_clear(struct mpls_instruction*,int,enum mpls_direction_enum,void*);
extern int mpls_instruction_build(struct mpls_instruction_req*,
  struct mpls_instruction*,int, void*);
extern struct net_device *mpls_dev_by_labelspace(int);

extern int mpls_tunnel_init(void);
extern struct net_device* mpls_tunnel_locate(struct ifreq* ifr,int create);
extern void mpls_tunnel_unlink(struct net_device* dev);
extern void mpls_tunnel_link(struct net_device* dev);
extern int mpls_tunnel_info(char *buffer,char **start,off_t offset,int length);

extern void mpls_instruction_copy(struct mpls_instruction *out,
  struct mpls_instruction *in,int length);

void *mpls_create_if_info(void);
void mpls_delete_if_info(struct mpls_interface* mpls_if);
void mpls_set_nexthop(struct dst_entry*, u32);

extern int mpls_version_info(char *buffer,char **start,off_t offset,int length);

#endif
